@@ -39,6 +39,7 @@
 #include <string.h>
 #include <ctype.h>
 #include <math.h>
+#include <sys/ioctl.h>
 #if HAVE_UNISTD_H || CRAENV
 #include <unistd.h>
 #endif
@@ -55,6 +56,8 @@
 #endif /* TIME_WITH_SYS_TIME */
 #include <errno.h>
 
+#include <sys/mman.h>
+
 #ifdef HAVE_WINSOCK2_H
 #include <winsock2.h>
 #endif
@@ -157,6 +160,10 @@ typedef u_int32_t socklen_t;
 #  endif
 #endif
 
+#ifdef HAVE_LIBJACK
+#include <jack/jack.h>
+#endif
+
 volatile sig_atomic_t stopit;
 volatile sig_atomic_t signal_report;
 volatile sig_atomic_t n_sighups;
@@ -211,6 +218,8 @@ struct progstate {
   int using_o_direct;
   int using_o_direct_i; /* Input */
   struct pmalloc pmalloc; /* Size is b (blocksize) */
+  off_t outputfile_position;
+  char *outputfile_map;
 };
 
 #ifndef NOTIMER 
@@ -277,7 +286,10 @@ sigreport(int signal)
 static void
 sighup(int signal)
 {
-  int meh = write(2, "SIGHUP\n", sizeof("SIGHUP\n") - 1);
+  const ssize_t meh = write(2, "SIGHUP\n", sizeof("SIGHUP\n") - 1);
+  if (meh != sizeof("SIGHUP\n")) {
+    /* ignore write error */
+  }
   n_sighups++;
 }
 
@@ -335,6 +347,51 @@ tsignal(const int sig, void (*const handler)(int))
 
 #ifdef WANT_SOUND
 
+#ifdef HAVE_LIBJACK
+
+int jack_callback(jack_nframes_t nframes, void *arg)
+{
+  //fprintf(stderr, "Jackd wants %d samples\n", nframes);
+  return 0;
+}
+
+void open_jack()
+{
+  jack_client_t *client;
+  jack_status_t status;
+  // TODO: use JackSessionID
+  client = jack_client_open("cstream", JackNoStartServer, &status);
+  if (client == 0) {
+    fprintf(stderr, "Failed to open jackd 0x%X\n", status);
+    exit(2);
+  }
+  jack_set_process_callback(client, jack_callback, 0);
+  jack_port_t *port1;
+  port1 = jack_port_register(client, "in_1", JACK_DEFAULT_AUDIO_TYPE
+			    , JackPortIsOutput, 0);
+  if (port1 == NULL) {
+    fprintf(stderr, "can't register jack port\n");
+    exit(2);
+  }
+  jack_port_t *port2;
+  port2 = jack_port_register(client, "in_2", JACK_DEFAULT_AUDIO_TYPE
+			    , JackPortIsOutput, 0);
+  if (port2 == NULL) {
+    fprintf(stderr, "can't register jack port\n");
+    exit(2);
+  }
+
+  if (jack_activate (client)) {
+    fprintf(stderr, "cannot activate jack client");
+    exit(2);
+  }
+
+  jack_native_thread_t jack_thread;
+  //jack_client_create_thread(client, &jack_thread, 0, 0, 
+
+}
+#endif
+
 static void
 setaudio(int fd, const char *spec)
 {
@@ -592,7 +649,7 @@ open_tcp(const struct options *const o, int mode)
         }
         char name[8192];
         char addr_string[INET6_ADDRSTRLEN];
-        // .... // FIXME - use inet_ntop
+        // FIXME - use inet_ntop
         getnameinfo(use_this_serv_addr, use_this_serv_addr_len,
                     name, sizeof(name), NULL, 0, 0);
         if (inet_ntop(result->ai_family, &use_this_serv_addr, addr_string, 
@@ -628,7 +685,7 @@ open_tcp(const struct options *const o, int mode)
     } *it, funcs[3] = {
       {getsockname, "getsockname", "Local binding"},
       {getpeername, "getpeername", "Remote binding"},
-      {NULL}
+      {NULL, NULL, NULL}
     };
     union {
       struct sockaddr sa;
@@ -743,6 +800,74 @@ void open_output_file(const struct options *const o
     perror("Cannot open output file");
     exit(2);
   }
+
+  if (strchr(o->O, 'M')) { // use mmap instead of write, for hugetlbfs
+    if (o->i == NULL || o->i[0] == '\0')
+      fprintf(stderr, "mmap output requires that I know input file in advance\n");
+
+    if (flags != (O_RDWR | O_CREAT | O_TRUNC)) {
+      fprintf(stderr, "mmaped output mode expected open(2) flags %d, got %d (continuing anyway)\n", O_RDWR | O_CREAT | O_TRUNC, flags);
+    }
+
+    struct stat statbuf;
+    if (fstat(state->ifd, &statbuf) == -1) {
+      fprintf(stderr, "mmap output requires that I know input file size for '%s' fd %d", o->i, state->ifd);
+      perror("stat");
+      exit(2);
+    }
+    off_t filesize = statbuf.st_size;
+
+    if (filesize % (2 * 1024 * 1024) != 0) {
+      filesize = filesize / (2 * 1024 * 1024) + 1;
+      filesize = filesize * 2 * 1024 * 1024;
+    }
+
+    if (filesize == 0) {
+      fprintf(stderr, "output file mmap mode refusing to work on zero size file '%s'\n", o->i);
+    }
+    if (o->v > 1) {
+      fprintf(stderr, "Using mmap for output file size %lld, open flags 0x%x\n", (long long)filesize, flags);
+    }
+#define USE_TRUNCATE
+
+#ifdef USE_LSEEK
+    if (lseek(state->ofd, filesize, SEEK_SET) == -1) {
+      fprintf(stderr, "Something went wrong seeking to end for output file size %lld\n", (long long)filesize, o->o);
+      perror("lseek");
+      exit(2);
+    }
+#endif
+    
+#ifdef USE_WRITE
+    if (lseek(state->ofd, filesize - 1, SEEK_SET) == -1) {
+      perror("seek");
+    }
+    char buf[2] = "1";
+    if (write(state->ofd, buf, 1) == -1) {
+      perror("write");
+      exit(2);
+    }
+#endif
+#ifdef USE_TRUNCATE
+    if (ftruncate(state->ofd, filesize) == -1) {
+      perror("ftruncate for mapped output file (continuing)");
+    }
+#endif
+    state->outputfile_position = 0;
+    if ((state->outputfile_map = 
+	 mmap(NULL, filesize, PROT_WRITE, MAP_SHARED, state->ofd, 0))
+	== MAP_FAILED) {
+      perror("mmap output file");
+      exit(2);
+    }
+    if (o->v > 2) {
+      fprintf(stderr, "mmaped from %p to %p, trying to write a byte... [expect \"worked\"]\n", state->outputfile_map, state->outputfile_map + filesize);
+    }
+    *(state->outputfile_map) = '\0';
+    if (o->v > 2) {
+      fprintf(stderr, "... worked\n");
+    }
+  }
 }
 
 static void
@@ -752,6 +877,7 @@ init(struct options *const o, struct progstate *const state
   struct timeval t;
 
   bzero(state, sizeof(*state));
+  state->outputfile_position = -1; // signals that we just write(2)
 
   pagesize = getpagesize();
 
@@ -771,13 +897,18 @@ init(struct options *const o, struct progstate *const state
 
   if (blocksize == 0) {
     state->b = 8192;
+    if (o->B == 0)
+      o->B = state->b;
   } else {
     state->b = blocksize;
     o->user_specified_blocksize = 1;
   }
 
+#if 0
+  // WTF?
   if (state->b == 0)
     state->b = state->b;
+#endif
 
   if (strchr(o->I, 'D'))
     state->using_o_direct_i = 1;
@@ -788,9 +919,9 @@ init(struct options *const o, struct progstate *const state
   if (o->w == 0)
     o->w = state->b;
 
-  if (state->b < state->b) {
+  if (o->B && o->B < state->b) {
     fprintf(stderr, "-B must not be lower than -b or -n (%d/%d)\n"
-	    , state->b, state->b);
+	    , o->B, state->b);
     exit(1);
   }
 
@@ -947,15 +1078,17 @@ init(struct options *const o, struct progstate *const state
 }
 
 static void
-closefiles(struct progstate *const state)
+closefiles(const struct options *const o, struct progstate *const state)
 {
   if (state->ofd != 1 && state->ofd != 2 && state->ofd != -1)
     if (close(state->ofd) == -1) {
       perror("Cannot close outfile");
+      exit(2);
     }
   if (state->ifd != 0 && state->ifd != -1)
     if (close(state->ifd) == -1) {
       perror("Cannot close infile");
+      exit(2);
     }
 }
 
@@ -1021,8 +1154,11 @@ report(const struct options *const o,
   print_kmg(" ", "%.2f", rate, "B/s", stderr);
   if (o->l)
       fprintf(stderr, " %g lines", (double)state->n_lines);
+#if 0
+  // WTF?
   if (curbytes != -1 && state->b != state->b)
     fprintf(stderr, " %.1f %%buf", (double)curbytes / (double)state->b * 100.0);
+#endif
   fprintf(stderr, "\n");
 
   if (o->v >= 2 && (sofar - state->time_firsttransfer) > 0.0) {
@@ -1082,6 +1218,20 @@ my_write(const struct options *const o, struct progstate *const state
     buf = state->pmalloc.data;
   }
 
+  // performance fixme.  It would be better to directly read(2) into
+  // this mmap'ed space, however the read code in here has all kinds
+  // of other complications so copy instead
+  if (state->outputfile_position != -1) {
+    if (o->v > 2) {
+      // ....
+      fprintf(stderr, "in mmap output file mode: copying %lld bytes, rel position %lld, addr %p\n", (long long)n_bytes, (long long)state->outputfile_position, state->outputfile_map + state->outputfile_position);
+    }
+    bcopy(buf, state->outputfile_map + state->outputfile_position, n_bytes);
+    state->outputfile_position += n_bytes;
+    return n_bytes;
+  }
+
+  // regular write
   ret = write(state->ofd, buf, n_bytes);
   if (ret != -1)
     return ret;
@@ -1261,11 +1411,17 @@ pollloop(const struct options *const o
       stopit = 1;
     }
   }
+  if (strchr(o->O, 'F')) {
+    if (fsync(state->ofd) < 0) {
+      perror("fsync");
+      exit(2);
+    }
+  }
+
+  closefiles(o, state);
   if (o->v >= 1)
     report(o, state, -1);
 
-  closefiles(state);
-
   free(palloc.allocated);
 }
 #endif /* HAVE_POLL_H */
@@ -1365,7 +1521,7 @@ loop(struct options *const o, struct progstate *const state)
 	  exit(1);
 	}
 
-	state->b = state->b = frames * 2 * 2; /* 16 bit Stereo */
+	state->b = frames * 2 * 2; /* 16 bit Stereo */
 	for (i = 0; i < frames; i++) {
 	  val = sin((double)i * 2.0 * pi / (double)frames) * 32767.0;
 	  buf[i * 4] = buf[i * 4 + 2] = val % 256;
@@ -1588,10 +1744,17 @@ loop(struct options *const o, struct progstate *const state)
     }
   }
 
+  if (strchr(o->O, 'F')) {
+    if (fsync(state->ofd) < 0) {
+      perror("fsync");
+      exit(2);
+    }
+  }
+
+  closefiles(o, state);
   if (o->v >= 1 && ! (o->c > 0 && childpid == 0)) /* Parent only */
     report(o, state, -1);
 
-  closefiles(state);
   if (o->c > 0) {
     if (childpid == 0) {
       exit(0);
@@ -1649,12 +1812,15 @@ usage(void)
 	  "-I <s> = Type of input file\n"
 	  "-O <s> = Type of ouput file\n"
           "         'f' = fifo (create it)\n"
+          "         'F' = issue fsync(2) on output file before closing\n"
           "         'a' = set audio modes on file (i.e. CD quality)\n"
           "         'N' = don't use TCP even if filename has ':'\n"
           "         't' = tee - in addition to outfile, "
 	  "copy stream to fd 3\n"
           "         'D' = O_DIRECT\n"
           "         'S' = O_SYNC\n"
+          "         'M' = use mmap for output file (do not use write(2)), "
+	  "for hugetlbfs\n"
 	  "         [Multiple chars allowed]\n"
 	  "-p <s> = Write pid as ascii integer to file <s>\n"
 	  "-l       include line count in statistics\n"
@@ -1700,6 +1866,8 @@ main(int argc, char *const argv[])
 
   default_options(&o);
 
+  //open_jack();
+
   while ((ch = getopt(argc, argv, "b:B:c:i:I:n:o:O:p:St:T:v:Vl6:")) != -1) {
     switch(ch) {
     case 'v': o.v = atoi(optarg); break;
